from enum import IntEnum
import os
import binascii
from iroha import IrohaCrypto
from iroha import Iroha, IrohaGrpc
import sys

import json
import re
from concurrent.futures import ThreadPoolExecutor

from google.protobuf.json_format import MessageToDict

class GeoPos:
  def __init__(self, lat, lon, time):
    self._lat = lat
    self._lon = lon
    self._time = time

  def get_lat(self):
    return self._lat

  def get_lon(self):
    return self._lon

  def get_time(self):
    return self._time

  def __str__(self):
    return "({}, {}) @ {}".format(self._lat, self._lon, self._time)

AWS = '13.53.243.73'
LOCAL_HOST = '127.0.0.1'

IROHA_HOST_ADDR = os.getenv('IROHA_HOST_ADDR', AWS)
IROHA_PORT = os.getenv('IROHA_PORT', '50051')

# ACCOUNT_ID = sys.argv[1]
ACCOUNT_ID = "200100250b3fa14c422d977815e6912e@0b3fa14c"
ACCOUNT_PRIVATE_KEY = ''
ACCOUNT_PUBLIC_KEY = ''

BASE_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")

with open("{}.priv".format("{}/identities/admin/{}".format(BASE_DIR, ACCOUNT_ID.split('@')[0])), 'r') as f:
  ACCOUNT_PRIVATE_KEY = f.read()

with open("{}.pub".format("{}/identities/admin/{}".format(BASE_DIR, ACCOUNT_ID.split('@')[0])), 'r') as f:
  ACCOUNT_PUBLIC_KEY = f.read()

iroha = Iroha(ACCOUNT_ID)
net = IrohaGrpc('{}:{}'.format(IROHA_HOST_ADDR, IROHA_PORT))

def trace(func):
  """
  A decorator for tracing methods' begin/end execution points
  """

  def tracer(*args, **kwargs):
    name = func.__name__
    print('\tEntering "{}"'.format(name))
    result = func(*args, **kwargs)
    print('\tLeaving "{}"'.format(name))
    return result

  return tracer

@trace
def send_transaction_and_print_status(transaction):
  hex_hash = binascii.hexlify(IrohaCrypto.hash(transaction))
  print('Transaction hash = {}, creator = {}'.format(
    hex_hash, transaction.payload.reduced_payload.creator_account_id))
  net.send_tx(transaction)
  for status in net.tx_status_stream(transaction):
    print(status)

@trace
def get_account_assets(account_id):
  """
  List all the assets of userone@domain
  """
  query = iroha.query('GetAccountAssets', account_id=account_id)
  IrohaCrypto.sign_query(query, ACCOUNT_PRIVATE_KEY)

  response = net.send_query(query)
  data = response.account_assets_response.account_assets
  for asset in data:
    print('Asset id = {}, balance = {}'.format(
      asset.asset_id, asset.balance))


@trace
def create_account(name, domain):
  with open("{}.pub".format(name + '@' + domain), 'r') as f:
    key = f.read()
  transaction = iroha.transaction([
    iroha.command(
      'CreateAccount',
      account_name = name,
      domain_id = domain,
      public_key = key
    )
  ])
  IrohaCrypto.sign_transaction(transaction, ACCOUNT_PRIVATE_KEY)
  net.send_tx(transaction)

  for status in net.tx_status_stream(transaction):
    print(status)

@trace
def set_account_details(account_id, key, val):
  tx = iroha.transaction([
    iroha.command(
      "SetAccountDetail",
      account_id=account_id,
      key=key,
      value=val
    )
  ])
  IrohaCrypto.sign_transaction(tx, ACCOUNT_PRIVATE_KEY)
  net.send_tx(tx)

  for status in net.tx_status_stream(tx):
    print(status)

@trace 
def get_account(account_id):
  query = iroha.query("GetAccount", account_id=account_id)
  IrohaCrypto.sign_query(query, ACCOUNT_PRIVATE_KEY)

  response = net.send_query(query)
  print(response)
  return json.loads(response.account_response.account.json_data)

@trace
def get_blocks():
  query = iroha.blocks_query()
  IrohaCrypto.sign_query(query, ACCOUNT_PRIVATE_KEY)
  blocks = net.send_blocks_stream_query(query, timeout=120)

  print(next(blocks))

@trace
def get_block(height):
  query = iroha.query("GetBlock", height=height)
  IrohaCrypto.sign_query(query, ACCOUNT_PRIVATE_KEY)

  response = MessageToDict(net.send_query(query))
  if response.get("errorResponse") != None:
    print("Reason: {}\nMessage: {}".format(response["errorResponse"]["reason"], response["errorResponse"]["message"]))
    return False, response["errorResponse"]["message"]
  else:
    return True, response["blockResponse"]["block"]["blockV1"]["payload"]


def find_block(height):
  query = iroha.query("GetBlock", height=height)
  IrohaCrypto.sign_query(query, ACCOUNT_PRIVATE_KEY)

  response = MessageToDict(net.send_query(query))
  if response.get("errorResponse") != None:
    print("Reason: {}\nMessage: {}".format(response["errorResponse"]["reason"], response["errorResponse"]["message"]))
    return None
  else:
    return response["blockResponse"]["block"]["blockV1"]["payload"]


def find_geo_pos(geo_locations, height, account_id):
  """
  Helper function for finding blocks using threads.

  @param geo_locations: global list in main thread,
  where worker threads will append geo locations.
  """

  query = iroha.query("GetBlock", height=height)
  IrohaCrypto.sign_query(query, ACCOUNT_PRIVATE_KEY)

  response = MessageToDict(net.send_query(query))
  if response.get("errorResponse") != None:
    return
  else:
    block = response["blockResponse"]["block"]["blockV1"]["payload"]

  for tx in block["transactions"]:
    for command in tx["payload"]["reducedPayload"]["commands"]:
      if not command.get("setAccountDetail"):
        continue

      cmd = command["setAccountDetail"]
      if not (cmd["key"] == "location" and cmd["accountId"] == account_id): #and cmd["flightId"] == flight_id):
        continue

      loc = json.loads(cmd["value"].replace('\\\"', '\"'))
      geo_locations.append(GeoPos(loc["lat"], loc["lon"], tx["payload"]["reducedPayload"]["createdTime"]))


def find_geo_path(account_id=ACCOUNT_ID, flight_id=0):
  """
  Return a list of all geo location updates for a given account,
  and flight id if specified.
  """

  # Ugly hack to get current block size of blockchain:
  # The error message of a GetBlock query will tell us the current size
  # of the blockchain. To get the size, just try query for a block
  # that does not exist.
  height = int(re.findall(r'(?<=\()\d+(?=\))', get_block(sys.maxsize)[1])[1])

  # get all geo locations using threads.
  geo_locations = []
  with ThreadPoolExecutor(max_workers=100) as executor:
    for i in range(1, height+1):
      executor.submit(find_geo_pos, geo_locations, i, account_id)

  geo_locations.sort(key=lambda loc: loc.get_time())

  return geo_locations